#ifndef TYPON_FUNDAMENTAL_DEQUE_HPP_INCLUDED
#define TYPON_FUNDAMENTAL_DEQUE_HPP_INCLUDED

#include <cstdint>
#include <memory>
#include <mutex>
#include <optional>
#include <type_traits>
#include <utility>

#include <typon/fundamental/ring_buffer.hpp>


namespace typon::fdt
{

  template <typename T>
  struct deque
  {
    using array_type = ring_buffer<T>;

    using u8 = typename ring_buffer<T>::u8;
    using u64 = typename ring_buffer<T>::u64;

    u64 _top {0};
    u64 _bottom {0};
    array_type * _array;
    std::mutex _mutex;

    deque(u8 bits = 3) noexcept
      : _array(new array_type(bits))
    {}

    ~deque()
    {
      delete _array;
    }

    void push(T x) noexcept
    {
      std::lock_guard<std::mutex> lock(_mutex);
      if (_bottom - _top > _array->capacity() - 1)
      {
        _array = _array->grow(_top, _bottom);
      }
      _array->put(_bottom, x);
      _bottom++;
    }

    std::optional<T> pop() noexcept
    {
      std::lock_guard<std::mutex> lock(_mutex);
      std::optional<T> x {};
      if (_top < _bottom)
      {
        _bottom--;
        x = _array->get(_bottom);
      }
      if (_array->capacity() > (_bottom - _top) * 4)
      {
        if (auto next = _array->shrink(_top, _bottom))
        {
          delete _array;
          _array = next;
        }
      }
      return x;
    }

    std::optional<T> steal() noexcept
    {
      std::lock_guard<std::mutex> lock(_mutex);
      std::optional<T> x {};
      if (_top < _bottom)
      {
        x = _array->get(_top);
        _top++;
      }
      return x;
    }
  };

}


#endif // TYPON_FUNDAMENTAL_DEQUE_HPP_INCLUDED
